﻿/*
 * Erstellt mit SharpDevelop.
 * Benutzer: Peter Zimmermann
 * Datum: 28.01.2011
 * Zeit: 23:46
 * 
 * Sie können diese Vorlage unter Extras > Optionen > Codeerstellung > Standardheader ändern.
 */
using System;
using System.Collections.Generic;
using System.Drawing;
using System.Windows.Forms;
using System.Net;
using System.Net.Sockets;

namespace TCP_Client
	{
	/// <summary>
	/// Description of MainForm.
	/// *********************************************************
	/// Demo-Programm in C#, kompiliert mit SharpDevelop IDE.
	/// Die hier gezeigte Programmierung erhebt keinen Anspruch
	/// auf "saubere, gut strukturierte" Vorgehensweise. Es soll 
	/// lediglich einen möglichen Lösungsansatz für die Kommuni-
	/// kation mit dem MFR-Modul über eine Netzwerkverbindung im
	/// TCP-Protokoll aufzeigen. Da die, in der Standard Library,
	/// verfügbaren Sockets blockierend sind, muss etwas Aufwand
	/// betrieben werden um ein "Hängenbleiben" der Applikation
	/// zu vermeiden. Da die Wahl von C# als Programmiersprache
	/// im Ermessen des Anwenders liegt und möglicherweise auch
	/// durch die Art der Applikation geprägt ist, kann lediglich
	/// auf den nicht unerheblichen (Schreib-)Aufwand in C# hin-
	/// gewiesen werden. Eine Wertung findet daher nicht statt.
	/// Die Kommentierung der einzelnen Schritte beschränkt sich
	/// auf das Wesentliche. Unkommentiertes muss ggf. aus den
	/// vewendeten Namen abgeleitet werden. Zur Realisierung
	/// wurden auch "Code-Schnipsel" aus Veröffentlichungen im
	/// Internet von unterschiedlichen Urhebern verwendet. Das
	/// Exception-Handling ruft in der Regel eine separate
	/// MessageBox auf, es wurde bewusst auf die Integration in
	/// die Demo-Applikation verzichtet. 
	///  
	/// Getestet wurde dieser Demo-Viewer auf Win-XP und Win 7. 
	/// *********************************************************
	/// </summary>
	public partial class MainForm : Form
		{
		IAsyncResult a_result;
		int current_out_bits;
		CheckBox[] cbO = new CheckBox[8];
		CheckBox[] cbI = new CheckBox[8];
		public AsyncCallback pfnCallBack;
		public Socket clientSocket;
		public System.String szReceivedData;
		
		public MainForm()
			{
			//
			// The InitializeComponent() call is required for Windows Forms designer support.
			//
			InitializeComponent();
			
			//
			// TODO: Add constructor code after the InitializeComponent() call.
			//
			lbLog.Items.Clear();
			cbO[0] = cbO0;
			cbO[1] = cbO1;
			cbO[2] = cbO2;
			cbO[3] = cbO3;
			cbO[4] = cbO4;
			cbO[5] = cbO5;
			cbO[6] = cbO6;
			cbO[7] = cbO7;
			cbI[0] = cbI0;
			cbI[1] = cbI1;
			cbI[2] = cbI2;
			cbI[3] = cbI3;
			cbI[4] = cbI4;
			cbI[5] = cbI5;
			cbI[6] = cbI6;
			cbI[7] = cbI7;
			
			current_out_bits = 0;
			UpdateControls(false);
			}
		
		byte[] StringToByteArray(string str)
			{
    		System.Text.ASCIIEncoding enc = new System.Text.ASCIIEncoding();
    		return enc.GetBytes(str);
			}
			
		void ShowStatus(string text)
			{
			lblStatus.Text = text;
			}
		
		private delegate void AddLogItemDelegate(object item);

		private void AddLogItem(object item)
			{
			if (lbLog.InvokeRequired)
				lbLog.Invoke(new AddLogItemDelegate(AddLogItem),item);
			else
				lbLog.Items.Add (item);
			}
			
		public class SocketPacket
			{
			public System.Net.Sockets.Socket thisSocket;
			public byte[] dataBuffer = new Byte[1];
			}
			
		public void WaitForData()
			{
			// init a new thread for receiving data from remote host
			try {
				if (pfnCallBack == null)
					{
					pfnCallBack = new AsyncCallback (OnDataReceived);
					}
				SocketPacket theSocPkt = new SocketPacket();
				theSocPkt.thisSocket = clientSocket;
				// Start listening ...
				a_result = clientSocket.BeginReceive (theSocPkt.dataBuffer,
																	0, theSocPkt.dataBuffer.Length,
																	SocketFlags.None,
																	pfnCallBack,
																	theSocPkt);
			} catch (Exception se) {
				MessageBox.Show (se.Message);
				}
			}
			
		public void OnDataReceived (IAsyncResult asyn)
			{
			try 
				{
				SocketPacket theSockId = (SocketPacket)asyn.AsyncState;
				int iRx = theSockId.thisSocket.EndReceive (asyn);
				char[] chars = new Char[iRx + 1];
				// whoever has sent any data,
				// we like ASCII ...
				System.Text.Decoder d = System.Text.Encoding.ASCII.GetDecoder();
				int charLen = d.GetChars(theSockId.dataBuffer, 0, iRx, chars, 0);
				for (int i=0; i<charLen; i++)
					{
					char c = chars[i];
					switch (c)
						{
						case '\r':	// CarriageReturn
									{
									// Command is complete, now process it ..
									// show it ..
									AddLogItem("RX <<: " + szReceivedData);
									ProcessCommand();
									szReceivedData = ""; // remove this command
									break;
									}
						case '\n':	// LineFeed
									{
									// skip it ..
									break;
									}
						default: {
									// collect received data
									if (c>0x1f && c<0x7f)
										szReceivedData += c;
									break;
									}
						}
					}
				WaitForData();
				}
			catch (ObjectDisposedException)
				{
				System.Diagnostics.Debugger.Log(0,"1","\nOnDataReceived: Socket has been closed\n");
				}
			catch (SocketException se)
				{
				MessageBox.Show (se.Message);
				}
			}
		
		void ProcessCommand()
			{
			// Processes the received message from the remote host
			int i_bits,o_bits;
			char cmd = szReceivedData[0];
			switch (cmd)
				{
				case 'I':
					// reflects the Inputs I0..7
					i_bits = (szReceivedData[1] & 0xf) << 4;
					i_bits += szReceivedData[2] & 0xf;
					InBits(i_bits);
					break;
				case 'O':
					// reflects the Outputs O0..7
					o_bits = (szReceivedData[1] & 0xf) << 4;
					o_bits += szReceivedData[2] & 0xf;
					current_out_bits = o_bits;
					OutBits (o_bits);
					break;
				// ignore all other ..
				}
			}
		
		void SetInBits (int ndx,int i_bits, int mask)
			{
			// note the different threads,
			// so we have to 'Invoke'!
			if (cbI[ndx].InvokeRequired)
				cbI[ndx].Invoke(new Action<int, int, int>(SetInBits), new object[] {ndx,i_bits,mask});
			else
				cbI[ndx].Checked = ((i_bits & mask) != 0)? true:false;
			}
			
		void InBits(int i_bits)
			{
			// Set or reset the output bits in a loop
			int mask = 1;
			for (int i=0; i<8; i++)
				{
				SetInBits(i,i_bits,mask);
				mask <<= 1;
				}
			}
		
		void SetOutBits (int ndx,int o_bits, int mask)
			{
			// note the different threads,
			// so we have to 'Invoke'!
			if (cbO[ndx].InvokeRequired)
				cbO[ndx].Invoke(new Action<int, int, int>(SetOutBits), new object[] {ndx,o_bits,mask});
			else
				{
				cbO[ndx].Checked = ((o_bits & mask) != 0)? true:false;
				}
			}
			
		void OutBits(int o_bits)
			{
			// Set or reset the output bits in a loop
			int mask = 1;
			for (int i=0; i<8; i++)
				{
				SetOutBits(i,o_bits,mask);
				mask <<= 1;
				}
			}
			
		void UpdateControls(bool connected)
			{
			// Enables or disables the controls
			btnConnect.Enabled = !connected;
			btnDisconnect.Enabled = connected;
			btnSend.Enabled = connected;
			btnInput.Enabled = connected;
			btnOutput.Enabled = connected;
			btnReset.Enabled = connected;
			for (int i=0; i<8; i++)
				{
				cbO[i].Enabled = connected;
				cbI[i].Enabled = connected;
				}
			string text = connected? "Connected" : "Not Connected";
			ShowStatus(text);
			AddLogItem("*** System: " + text);
			OutBits(0);
			InBits(0);
			}
		
		void BtnConnectClick(object sender, EventArgs e)
			{
			// connect to the remote server, if any ..
			if (txtBoxIP.Text == "" || txtBoxPort.Text == "") {
				// do anything
				ShowStatus("IP Address and/or Port Number required!");
				return;
				}
			try {
				UpdateControls (false);
				clientSocket = new Socket (AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
				IPAddress ip = IPAddress.Parse (txtBoxIP.Text);
				int iPortNo = System.Convert.ToInt32 (txtBoxPort.Text);
				IPEndPoint ipEnd = new IPEndPoint(ip, iPortNo);
				// connect to the remote host
				clientSocket.Connect (ipEnd);
				if (clientSocket.Connected) 
					{
					UpdateControls (true);
					// wait for data
					WaitForData();
					SendCommand("O");
					System.Threading.Thread.Sleep(20); // may be omitted
					SendCommand("I");
					}
			} catch (Exception) {
				// connection failed	
				UpdateControls (false);
				ShowStatus("Connection failed, Server not found!");
				}
			}
		
		void BtnDisconnectClick(object sender, EventArgs e)
			{
			// closes an open connection			
			if (clientSocket != null) {
				clientSocket.Close();
				clientSocket = null;
				UpdateControls(false);
				}
			}
		
		void BtnClearLogClick(object sender, EventArgs e)
			{
			lbLog.Items.Clear();
			}
			
		void SendCommand(string cmd)
			{
			// send command string to the remote host
			AddLogItem("TX >>: " + cmd);
			try {
				byte[] byData = StringToByteArray(cmd + "\r");
				if (clientSocket != null)
					{
					clientSocket.Send (byData);
					}
			} catch (Exception se) {
				MessageBox.Show (se.Message);	
				}
			}
		
		void BtnSendClick(object sender, EventArgs e)
			{
			// send customer issued command from textbox
			if (txtCommand.Text == "") return;
			// the command will not be checked ..
			SendCommand (txtCommand.Text);
			}
		
		void BtnResetClick(object sender, EventArgs e)
			{
			// Short command to reset the module
			// Attention: It is a 'Software Reset',
			// a working communication channel is required!
			OutBits(0);	// for visual effect only
			InBits(0);	// for visual effect only
			current_out_bits = 0;
			SendCommand ("X");
			}
		
		void BtnInputClick(object sender, EventArgs e)
			{
			// Short command for getting Inputs
			SendCommand ("I");
			}
		
		void BtnOutputClick(object sender, EventArgs e)
			{
			// Short command for getting Outputs
			SendCommand ("O");
			}
		
		void BtnCloseClick(object sender, EventArgs e)
			{
			// closes the application			
			if (clientSocket != null) {
				clientSocket.Close();
				clientSocket = null;
				}
			Close();
			}
		
	
		void CbO0Click(object sender, EventArgs e)
			{
			// Clicking one of the CheckBoxes toggles
			// this Output (on the Module).
			string cmd = "O";
			int bits = current_out_bits;
			int tag = System.Convert.ToInt16(((CheckBox)sender).Tag);
			((CheckBox)sender).Checked = !((CheckBox)sender).Checked;
			if (((CheckBox)sender).Checked)
				bits |= tag;
			else
				bits &= ~tag;
			cmd += (char)('@' + (bits >> 4));
			cmd += (char)('@' + (bits & 0xF));
			SendCommand(cmd);
			
			}
		}
	}
